import datetime
import pandas as pd
import os

from export_vdf.util import extract_data_hash


class ExportVDB:
    def __init__(self, args):
        self.args = args
        self.file_structure = []
        self.file_ctr = 1
        self.hash_value = extract_data_hash(self.args)
        self.timestamp_in_format = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        self.vdf_directory = f"vdf_{self.timestamp_in_format}_{self.hash_value}"
        os.makedirs(self.vdf_directory, exist_ok=True)

    def get_data():
        """
        Get data from vector database
        """
        raise NotImplementedError

    def save_vectors_to_parquet(self, vectors, metadata, batch_ctr, vectors_directory):
        vectors_df = pd.DataFrame(list(vectors.items()), columns=["id", "vector"])
        # Store the vector in values as a column in the parquet file, and store the metadata as columns in the parquet file
        # Convert metadata to a dataframe with each of metadata_keys as a column
        # Convert metadata to a list of dictionaries
        metadata_list = [{**{"id": k}, **v} for k, v in metadata.items()]
        # Convert the list to a DataFrame
        metadata_df = pd.DataFrame.from_records(metadata_list)
        # Now merge this metadata_df with your main DataFrame
        df = vectors_df.merge(metadata_df, on="id", how="left")

        # Save the DataFrame to a parquet file
        parquet_file = os.path.join(vectors_directory, f"{batch_ctr}.parquet")
        df.to_parquet(parquet_file)
        self.file_structure.append(parquet_file)

        # Reset vectors and metadata
        vectors = {}
        metadata = {}
        return len(df)



import os
from dotenv import load_dotenv
import pandas as pd
from qdrant_client import QdrantClient
from export_vdf.util import extract_numerical_hash
from import_vdf.vdf_import_cls import ImportVDF
from qdrant_client.http.models import VectorParams, Distance, PointStruct, UpdateStatus

load_dotenv()


class ImportQdrant(ImportVDF):
    def __init__(self, args):
        # call super class constructor
        super().__init__(args)
        self.client = QdrantClient(
            url=self.args["url"], api_key=self.args["qdrant_api_key"]
        )

    def upsert_data(self):
        # we know that the self.vdf_meta["indexes"] is a list
        for index_name, index_meta in self.vdf_meta["indexes"].items():
            # load data
            print(f"Importing data for index '{index_name}'")
            for namespace_meta in index_meta:
                data_path = namespace_meta["data_path"]
                # list indexes
                collections = [
                    x.name for x in self.client.get_collections().collections
                ]
                # check if index exists
                new_collection_name = index_name + (
                    f"_{namespace_meta['namespace']}"
                    if namespace_meta["namespace"]
                    else ""
                )
                if new_collection_name not in collections:
                    # create index
                    try:
                        self.client.create_collection(
                            collection_name=new_collection_name,
                            vectors_config=VectorParams(
                                size=namespace_meta["dimensions"],
                                distance=(
                                    namespace_meta["metric"]
                                    if "metric" in namespace_meta
                                    else Distance.COSINE
                                ),
                            ),
                        )
                    except Exception as e:
                        print(f"Failed to create index '{new_collection_name}'", e)
                        return
                prev_vector_count = self.client.get_collection(
                    collection_name=new_collection_name
                ).vectors_count
                if prev_vector_count > 0:
                    print(
                        f"Index '{new_collection_name}' has {prev_vector_count} vectors before import"
                    )
                # Load the data from the parquet files
                parquet_files = sorted(
                    [
                        file
                        for file in os.listdir(data_path)
                        if file.endswith(".parquet")
                    ]
                )

                vectors = {}
                metadata = {}
                vector_column_names, vector_column_name = self.get_vector_column_name(
                    new_collection_name, namespace_meta
                )
                for file in parquet_files:
                    file_path = os.path.join(data_path, file)
                    df = pd.read_parquet(file_path)
                    vectors.update(
                        {row["id"]: row[vector_column_name] for _, row in df.iterrows()}
                    )
                    metadata.update(
                        {
                            row["id"]: {
                                key: value
                                for key, value in row.items()
                                if key not in ["id"] + vector_column_names
                            }
                            for _, row in df.iterrows()
                        }
                    )
                vectors = {k: v.tolist() for k, v in vectors.items()}
                response = self.client.upsert(
                    collection_name=new_collection_name,
                    points=[
                        PointStruct(
                            id=int(idx)
                            if idx.isdigit()
                            else extract_numerical_hash(idx),
                            vector=vectors[idx],
                            payload=metadata.get(idx, {}),
                        )
                        for idx in vectors.keys()
                    ],
                    wait=True,
                )
                if response.status != UpdateStatus.COMPLETED:
                    print(
                        f"Failed to upsert data for collection '{new_collection_name}'"
                    )
                    continue
                vector_count = self.client.get_collection(
                    collection_name=new_collection_name
                ).vectors_count
                print(
                    f"Index '{new_collection_name}' has {vector_count} vectors after import"
                )
                print(f"{vector_count - prev_vector_count} vectors were imported")
        print("Data import completed successfully.")



import json
from qdrant_client import QdrantClient
import os
from tqdm import tqdm
from dotenv import load_dotenv

from export_vdf.vdb_export_cls import ExportVDB

load_dotenv()

MAX_FETCH_SIZE = 10_000


class ExportQdrant(ExportVDB):
    def __init__(self, args):
        """
        Initialize the class
        """
        super().__init__(args)
        try:
            self.client = QdrantClient(
                url=self.args["url"], api_key=self.args["qdrant_api_key"]
            )
        except:
            self.client = QdrantClient(url=self.args["url"])

    def get_all_collection_names(self):
        """
        Get all collection names from Qdrant
        """
        collections = self.client.get_collections().collections
        collection_names = [collection.name for collection in collections]
        return collection_names

    def get_data(self):
        if "collections" not in self.args or self.args["collections"] is None:
            collection_names = self.get_all_collection_names()
        else:
            collection_names = self.args["collections"].split(",")

        index_metas = {}
        for collection_name in tqdm(collection_names, desc="Fetching indexes"):
            index_meta = self.get_data_for_collection(collection_name)
            index_metas[collection_name] = index_meta

        # Create and save internal metadata JSON
        self.file_structure.append(os.path.join(self.vdf_directory, "VDF_META.json"))
        internal_metadata = {
            "version": self.args["library_version"],
            "file_structure": self.file_structure,
            "author": os.environ.get("USER"),
            "exported_from": "qdrant",
            "indexes": index_metas,
        }
        with open(os.path.join(self.vdf_directory, "VDF_META.json"), "w") as json_file:
            json.dump(internal_metadata, json_file, indent=4)
        # print internal metadata properly
        print(json.dumps(internal_metadata, indent=4))
        return True

    def try_scroll(self, fetch_size, collection_name, next_offset):
        try:
            records, next_offset = self.client.scroll(
                collection_name=collection_name,
                offset=next_offset,
                limit=fetch_size,
                with_payload=True,
                with_vectors=True,
            )
            return records, next_offset, fetch_size
        except:
            print("Failed to fetch data, reducing fetch size to", (fetch_size * 2) // 3)
            return self.try_scroll((fetch_size * 2) // 3, collection_name, next_offset)

    def get_data_for_collection(self, collection_name):
        vectors_directory = os.path.join(self.vdf_directory, collection_name)
        os.makedirs(vectors_directory, exist_ok=True)

        vectors = {}
        metadata = {}
        total = self.client.get_collection(collection_name).vectors_count

        num_vectors_exported = 0
        dim = self.client.get_collection(collection_name).config.params.vectors.size
        next_offset = 0
        records, next_offset, fetch_size = self.try_scroll(
            MAX_FETCH_SIZE, collection_name, next_offset
        )
        num_vectors_exported += self.save_from_records(
            vectors,
            metadata,
            records,
            vectors_directory,
        )
        pbar = tqdm(total=total, desc=f"Exporting {collection_name}")
        while next_offset is not None:
            records, next_offset, fetch_size = self.try_scroll(
                fetch_size, collection_name, next_offset
            )
            num_vectors_exported += self.save_from_records(
                vectors,
                metadata,
                records,
                num_vectors_exported,
                vectors_directory,
            )
            pbar.update(len(records))

        namespace_meta = {
            "index_name": collection_name,
            "namespace": "",
            "total_vector_count": total,
            "exported_vector_count": num_vectors_exported,
            "metric": self.client.get_collection(
                collection_name
            ).config.params.vectors.distance,
            "dimensions": dim,
            "model_name": self.args["model_name"],
            "vector_columns": ["vector"],
            "data_path": vectors_directory,
        }

        return {"": [namespace_meta]}

    def save_from_records(self, vectors, metadata, records, vectors_directory):
        num_vectors_exported = 0
        for point in records:
            vectors[point.id] = point.vector
            metadata[point.id] = point.payload
        num_vectors_exported += self.save_vectors_to_parquet(
            vectors, metadata, self.file_ctr, vectors_directory
        )
        return num_vectors_exported
